/******************************************************************************
    (c) 2002      Christine Caulfield          christine.caulfield@googlemail.com

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.
 ******************************************************************************
 */

/* Key definitions */
#define CTRL_A	0x01
#define CTRL_B	0x02
#define CTRL_C	0x03
#define CTRL_D	0x04
#define CTRL_E	0x05
#define CTRL_F	0x06
#define CTRL_G	0x07
#define BELL	0x07
#define CTRL_H	0x08
#define BS	0x08
#define CTRL_I	0x09
#define CTRL_J	0x0A
#define CTRL_K	0x0B
#define CTRL_L	0x0C
#define CTRL_M	0x0D
#define CTRL_N	0x0E
#define CTRL_O	0x0F
#define CTRL_P	0x10
#define	CTRL_Q  0x11
#define CTRL_R	0x12
#define CTRL_S	0x13
#define	CTRL_T	0x14
#define CTRL_U	0x15
#define CTRL_V	0x16
#define CTRL_W	0x17
#define	CTRL_X	0x18
#define CTRL_Y	0x19
#define CTRL_Z	0x1A
#define ESC	0x1B
#define DEL	0x7F

